/*
<trigger>
	<usage>bsf transition.js -to state_name [-rel relationship_name] [-ifattr attribute_name=value[,...][:...]]</usage>
	<version>5.0</version>
	<description>
		将同一个 CR、相关 CR 或关联任务转变为指定状态。</description>
	<argument1>
		<name>-to state_name</name>
		<required>yes</required>
		<type>string</type>
		<description>
			将对象转变为的状态名称。</description>
	</argument1>
	<argument2>
		<name>-rel relationship_name</name>
		<required>no</required>
		<type>string</type>
		<description>
			指定针对哪个对象执行转变。如果省略，那么将对当前 CR 执行转变。
			如果与“associated_task”一起使用，那么将对所有关联任务执行转变。对于任何其他值，将对与当前 CR 相关且具有关系“relationship_name”的 CR 执行转变。
		</description>
	</argument2>
	<argument3>
		<name>-ifattr attribute_name=value[,...][:...]</name>
		<required>no</required>
		<type>string</type>
		<description>
			如果目标对象满足属性条件“attribute_name=value”，那么执行转变。可以使用“=”号或“!=”号将属性条件指定为等于或不等于。
			例如，“-ifattr CRPR_phase!=resolution”表示如果属性“CRPR_phase”不等于“resolution”，那么执行转变。
			可以将多个值指定为以“,”字符分隔的列表。例如，“-ifattr CRPR_phase!=resolution,evaluation”表示如果属性“CRPR_phase”既不等于“resolution”也不等于“evaluation”，那么执行转变。
			可以将多个条件指定为以冒号分隔的列表。例如，“-ifattr CRPR_phase!=resolution:state=completed”表示如果属性“CRPR_phase”不等于“resolution”并且属性“state”等于“completed”，那么执行转变。
		</description>
	</argument3>
</trigger>
*/


// Import packages
importPackage(Packages.com.telelogic.cs.api);

// Script specific global variables
const script_name = 'transition.js';
const script_version = '20080606';
const script_usage = 'transition.js -to state_name [-rel relationship_name] [-ifattr attribute_name=value[,...][:...]]';
const script_example = 'transition.js -to task_assigned -rel associated_task -ifattr state=task_deferred:CRPR_phase!=resolution,evaluation';
const log_header = 'Script bsf ' + script_name + ': ';
const script_debug = false;
const script_dcm_modifiable_check = true;

// Get objects
var args = bsf.lookupBean('args');
var event = bsf.lookupBean('event');
var change = bsf.lookupBean('api');
var log = bsf.lookupBean('log');
var options = OptionParser.parse(args);

if (script_debug) 
{
	log.info(log_header);
	log.info('    Version: ' + script_version);
	log.info('    Arguments: ' + args.join(' '));
	log.info('    Current CR: ' + event.objectId);
}

if (checkArguments(0, ['-to', '-rel', '-ifattr'], ['-to']))
{

	// Get Options
	const to = options.getOption('-to');
	const rel = options.getOption('-rel');
	const ifattr = options.getOption('-ifattr');

	// Parse -ifattr
	var conditions = parseConditions(ifattr);

	// Create attribute list from the conditions
	var attribute_names = getAttributeNamesFromConditions(conditions);
	
	if (script_dcm_modifiable_check)
	{
		attribute_names.push('modifiable_in');
	}
	
	if (script_debug)
	{
		log.info(log_header + 'Condition attribute list: ' + attribute_names.join(', '));
	}

	// Branch over relationship type
	var items;
	
	if (!rel)
	{
		// Self
		if (script_debug) 
		{
			log.info(log_header + 'Working on self');
		}
		items = change.query(event.triggerUser.token, change.buildCrQuery(event.triggerUser.token, event.objectId), attribute_names);
	}
	else if (rel == 'associated_task')
	{
		// Associated tasks
		if (script_debug) 
		{
			log.info(log_header + 'Working on associated tasks');
		}
		items = change.getAssociatedTasks(event.triggerUser.token, event.objectId, attribute_names);
	}
	else
	{
		// Associated CRs
		if (script_debug) 
		{
			log.info(log_header + 'Working on associated CRs');
		}

		var has_string = 'has_'; 
		
		if (rel.indexOf(has_string) == 0)
		{
			items = change.getRelatedTo(event.triggerUser.token, event.objectId, rel.substr(has_string.length), attribute_names);
		}
		else
		{
			items = change.getRelatedFrom(event.triggerUser.token, event.objectId, rel, attribute_names);
		}
	}

	// Filter list
	if (script_debug) 
	{
		log.info(log_header + 'Item list: ' + items.join(', '));
	}

	items = filterItems(items, conditions);
	
	if (script_debug) 
	{
		log.info(log_header + 'Filtered item list: ' + items.join(', '));
	}

	// Action items
	if (script_debug) 
	{
		log.info(log_header + 'Item list: ' + items.join(', '));
	}

	for (var i = 0; i < items.length; i++)
	{
		transitionItem(items[i], to);
	}
}


/**
 * function checkArguments
 * 
 * Checks the script arguments:
 * - Check argument count
 * - Check that all specified options are valid
 * - Check that all required options are specified
 *
 * Returns: true, if succesfull. false otherwise
 */
function checkArguments(arg_count, allowed_options, required_options)
{
	var message;

	// Check argument count
	if (options.args.length < arg_count)
		message = 'Insufficient arguments to run the script (' + options.args.length +' arguments)';

	// Check that all specified options are valid
	if (!message)
	{
		var all_options = options.getOptions();
		var i = all_options.keySet().iterator();
		
		while (!message && i.hasNext())
		{
			var option = i.next();
			var is_valid = false;
		
			for (var j = 0; j < allowed_options.length; j++)
			{
				if (option == allowed_options[j])
				{
					is_valid = true;
					break;
				}
			}

			if (!is_valid)
			{
				message = 'Invalid option \'' + option + '\'';
			}
		}
	}

	// Check that all required options are specified
	if (!message) 
	{
		for (var i = 0; i < required_options.length; i++)
		{
			var option = options.getOption(required_options[i]);

			if (!option)
			{
				message = 'Missing option \'' + required_options[i] + '\'';
			}
		}
	}

	if (message)
	{
		log.error(log_header);
		log.error('   ' + message);
		log.error('   Usage: ' + script_usage);
		log.error('   Example: ' + script_example);

		return false;
	}

	return true;
}


/**
 *
 * Checks, if an attribute has no value set.
 *
 * The function is necessary, because Attribute.toString() returns 'null' instead of null. By
 * returning 'null' it cannot be distinguished between an without a value and one with the 
 * value 'null'.
 *
 * Returns: True, if the attriute has no value set.
*/
function isAttributeNull(attribute)
{
	if (((!attribute) || (attribute.type == Attribute.Type.STRING) && (attribute.getStringValue() == null))
		|| ((attribute.type == Attribute.Type.DATE) && (attribute.getDateValue() == null))
		|| ((attribute.type == Attribute.Type.BOOLEAN) && (attribute.getBooleanValue() == null)))
	{
		return true;
	}

	return false;
}


/**
 *
 * Parses the -ifattr string and creates a conditions definition from it.
 *
 * The conditions definition is an array of condition entries. Each
 * entry is an array with three elements:
 *   [0] Attribute name
 *   [1] Boolean indicating, if this is a not equal condition
 *   [2] Array with possible values. The array contains an additional 
 *       entry at the end to allow a '' element at the last position
 *
 * Returns: Condition definition, if succesfull. Null otherwise
 */
function parseConditions(ifattr)
{
	const error = log_header + 'Invalid -ifattr string \'' + ifattr + '\'';
	var conditions = new Array();

	if (ifattr)
	{
		var raw_conditions = ifattr.split(':');
		
		if (script_debug) 
		{
			log.info(log_header + 'Working on ' + raw_conditions.length + ' condition(s)');
		}

		for (var i = 0; i < raw_conditions.length; i++)
		{
			if (script_debug) 
			{
				log.info(log_header + 'i=' + i + ': ' + raw_conditions[i]);
			}

			var name;
			var is_not_equal;
			var values;

			// Split into name and values. 
			// Add '=X' to get empty value as null. Will always split to 3 parts then
			raw_conditions[i] = raw_conditions[i] + '=X';
			var raw_condition = raw_conditions[i].split('=');
			
			if (raw_condition.length != 3 || (raw_condition[0] == ''))
			{
				log.error(error);
				return null;
			}

			// Check for !=
			name = raw_condition[0];
			is_not_equal = (name.lastIndexOf('!') == name.length() - 1);

			// Name. Must be at least one character
			if (is_not_equal)
			{
				name = name.substr(0, name.length() - 1);
			}
			
			if (script_debug) 
			{
				log.info(log_header + 'Name: ' + name);
			}
			
			if (name.length == 0)
			{
				log.error(error);
				return null;
			}

			// Split into values
			// Add '=X' to get empty value at the end as null. Will always split to n+1 parts then
			raw_condition[1] = raw_condition[1] + ',X';
			values = raw_condition[1].split(',');
			
			if (values.length == 0)
			{
				log.error(error);

				return null;
			}
			
			if (script_debug)
			{
				log.info(log_header + 'Values: ' + (values.length - 1) );
			}

			var condition = [name, is_not_equal, values];
			conditions.push(condition);
		}
	}

	if (script_debug)
	{
		log.info(log_header + 'Processed ' + conditions.length + ' condition(s)');
	}

	return conditions;
}


/**
 *
 * Creates a list of attribute names based on the condition definition
 *
 * Returns: Array of attribute names
 */
function getAttributeNamesFromConditions(conditions)
{
	var attribute_names = new Array();

	for (var i = 0; i < conditions.length; i++)
	{
		attribute_names.push(conditions[i][0]);
	}

	return attribute_names;
}

/**
 * Filters items from the array of items to be handled by the script
 * based on the the conditions and if the item is modifiable in the 
 * current database
 *
 * Returns: Filtered list of items
 */
function filterItems(items, conditions)
{
	var new_items = new Array();

	for (var i = 0; i < items.length; i++)
	{
		if (filterItem(items[i], conditions))
		{
			new_items.push(items[i])
		}
	}

	return new_items; 
}

/**
 * Tests, if an item fullfills the specified conditions. The function
 * tests, if the item is modifiable in the current database and 
 * script_dcm_modifiable_check is true.
 *
 * Returns: True, if the item passes the test. False otherwise
 */
function filterItem(item, conditions)
{
	if (script_debug)
	{
		log.info(log_header + 'Testing item ' + item);
	}

	if (script_dcm_modifiable_check &&
		change.isDcmEnabled(event.triggerUser.token) &&
		!item.getAttribute('modifiable_in').toString().equals(change.getDbid(event.triggerUser.token))) // == is always false
	{
		if (script_debug) 
		{
			log.info(log_header + 'Skipping non modifiable item. Item \'' + item + 
			'\' is not modifiable in database \'' + change.getDbid(event.triggerUser.token) + '\'');
		}

		return false;
	}

	for (var i = 0; i < conditions.length; i++)
	{
		var condition = conditions[i];

		if (script_debug)
		{
			log.info(log_header + 'condition ' + i + ': Attribute ' + condition[0]);
		}
		
		// Get attribute and value. Use '' for no value
		var attribute = item.getAttribute(condition[0]);
		var attribute_value;
		
		if (!isAttributeNull(attribute))
		{
			attribute_value = attribute.toString();
		}
		else
		{
			attribute_value = '';
		}
		
		if (script_debug)
		{
			log.info(log_header + 'condition ' + i + ': Attribute value \'' + attribute_value + '\'');
		}

		// Get comparison mode
		var is_not_equal = condition[1];
		
		if (script_debug) 
		{
			log.info(log_header + 'condition ' + i + ': Is not equal: ' + is_not_equal);
		}
		
		// Compare by looping over values. Omit last entry. It is not a value
		var result = false;
	
		for (var j = 0; j < condition[2].length - 1; j++)
		{
			if (script_debug) 
			{
				log.info(log_header + 'condition ' + i + ': Comparing against value \'' + condition[2][j] + '\'');
			}

			// Compare
			var comparisson = (attribute_value.equals(condition[2][j])); // == is always false

			if (is_not_equal)
			{
				comparisson  = !comparisson;
			}
			
			if (script_debug) 
			{
				log.info(log_header + 'condition ' + i + ': Comparisson result ' + comparisson);
			}
			
				result = result || comparisson;
			}
			
			// Handle result
			if (script_debug)
			{
				log.info(log_header + 'condition ' + i + ': Test result ' + result);
			}

			if (!result)
			{
				return false;
			}
	}

	return true;
}

/**
 *
 * Transition CR or task
 *
 */
function transitionItem(item, to)
{
	var is_task = (item.toString().indexOf('Task') == 0);
	var attributes = new Array();
	var item_id_string;

	try 
	{
		if (is_task)
		{
			// Transition task

			item_id_string = 'Task ' +  item.taskId;
			if (script_debug) 
			{
				log.info(log_header + 'Transitioning ' + item_id_string + ' to state ' + to);
			}

			change.transitionTask(event.triggerUser.token, item.taskId, to, attributes);

		} 
		else 
		{
			// Transition CR
			item_id_string = 'CR ' +  item.crid;
			if (script_debug)
			{
				log.info(log_header + 'Transitioning ' + item_id_string + ' to state ' + to);
			}

			change.transition(event.triggerUser.token, item.crid, to, attributes);

		}
	}
	catch (error)
	{
		var message = change.getMessage('errorTransitioningItemXtoStateYerrorZ', [item_id_string, to, error.message || error.toString()]);		
		log.error(log_header + 'Copying from CR ' + event.objectId + '. ' + message); 
		event.fail(message);
	}
}
